/*
 * \brief  Client interface for VFS trust anchor access
 * \author Jean-Adrien Domage
 * \date   2022-07-05
 */

#ifndef _TRUST_ANCHOR__CLIENT_H_
#define _TRUST_ANCHOR__CLIENT_H_

#include <base/env.h>
#include <base/heap.h>
#include <os/path.h>
#include <vfs/simple_env.h>

#include <trust_anchor/types.h>

namespace Trust_anchor {
	using namespace Genode;
	using Path = Genode::Path<256>;

	constexpr const unsigned MAX_BUFF_SIZE = 4096;

	class Client;
}

/**
 * \attention Not thread safe
 */
class Trust_anchor::Client {
private:
	const Path        _path;
	Vfs::Simple_env  &_vfs_env;
	Vfs::File_system &_vfs;
	Genode::Heap     &_heap;

	size_t _write(Vfs::Vfs_handle *handle, Const_byte_range_ptr const &src) {
		size_t out_count;
		auto write_result = handle->fs().write(handle,
		                                       src,
		                                       out_count);
		if (write_result != Vfs::File_io_service::WRITE_OK) {
			Genode::error("TA::Client: write error");
			return 0;
		}
		return out_count;
	}

	size_t _read(Vfs::Vfs_handle *handle, Byte_range_ptr const &dst)
	{
		size_t out_count;
		auto read_result = handle->fs().complete_read(handle,
		                                dst,
		                                out_count);
		if (read_result != Vfs::File_io_service::READ_OK) {
			Genode::error("TA::Client: read error");
			return 0;
		}
		return out_count;
	}

	Vfs::Vfs_handle *_open(String<256> const &path, Vfs::Directory_service::Open_mode mode)
	{
		Vfs::Vfs_handle *file_handle;
		auto result = _vfs.open(path.string(),
		                        mode,
		                        &file_handle,
		                        _heap);
		if (result != Vfs::Directory_service::Open_result::OPEN_OK) {
			Genode::error("TA::Client: could not open ", path);
			return nullptr;
		}
		return file_handle;
	}

	size_t _cypher_service(String<256>  path,
	                      Const_byte_range_ptr const &src,
	                      Byte_range_ptr const &dst)
	{
		if (src.num_bytes > MAX_BUFF_SIZE) {
			Genode::error("TAClient: buffer is too large!");
			return 0;
		}
		State st = state();
		if (st != State::IDLE) {
			Genode::error("TAClient: hardware is not IDLE, current state is ",
			              state_to_string(st));
			return 0;
		}

		Vfs::Vfs_handle *file_handle = _open(path, Vfs::Directory_service::OPEN_MODE_RDWR);
		Vfs::Vfs_handle::Guard handle_guard(file_handle);
		if (!file_handle) return 0;

		size_t w_count = _write(file_handle, src);
		if (w_count != src.num_bytes) {
			Genode::error("TA::Client: i/o error in: ", src.num_bytes, " out: ", w_count);
			return 0;
		}

		if (state() == State::ERROR) {
			return 0;
		}
		return  _read(file_handle, dst);
	}

public:

	Client(Vfs::Simple_env &vfs_env, Genode::Heap &heap, Path const &path)
	:
		_path    { path },
		_vfs_env { vfs_env },
		_vfs     { _vfs_env.root_dir() },
		_heap    { heap }
	{
	}

	Client(const Client&) = delete;
	Client(Client&&) = delete;
	Client &operator=(const Client&) = delete;
	Client &operator=(Client&&) = delete;

	/**
	 * Read the state of the device.
	 */
	State state()
	{
		String<256>      path { _path, "/", Services::STATUS};
		Vfs::Vfs_handle *file_handle = _open(path, Vfs::Directory_service::Open_mode::OPEN_MODE_RDONLY);
		Vfs::Vfs_handle::Guard handle_guard(file_handle);
		if (!file_handle)
			return State::ERROR;
		State state;
		Byte_range_ptr dst { reinterpret_cast<char*>(&state), sizeof(state) };
		size_t size = _read(file_handle, dst);
		if (size == 0)
			return State::ERROR;
		return state;
	}

	/**
	 * Encrypt provided buff using the device.
	 * @param src       - buffer containing the data ro encrypt
	 * @param dst       - buffer to store the cyphered text
	 * @return The number of bytes wrote to out_buff.
	 */
	size_t encrypt(Const_byte_range_ptr const &src,
	               Byte_range_ptr       const &dst)
	{
		String<256> path { _path, "/", Services::ENCRYPT };
		return _cypher_service(path, src, dst);
	}

	/**
	 * Decrypt a cypher using the device.
	 * @param src       - buffer containing the cypher
	 * @param dst       - buffer to store the result in
	 * @return The number of bytes wrote to out_buf.
	 */
	size_t decrypt(Const_byte_range_ptr const &src,
	               Byte_range_ptr       const &dst)
	{
		String<256> path { _path, "/", Services::DECRYPT };
		return _cypher_service(path, src, dst);
	}

	/**
	 * Store hash to the device.
	 * \param src      - buff containing hash to store
	 * \return Number of bytes wrote to the device.
	 */
	size_t store_hash(Const_byte_range_ptr const &src)
	{
		String<256> path { _path, "/", Services::HASH_STORAGE };
		Vfs::Vfs_handle* file_handle = _open(path, Vfs::Directory_service::OPEN_MODE_WRONLY);
		if (file_handle == nullptr) return 0;
		Vfs::Vfs_handle::Guard handle_guard(file_handle);
		return _write(file_handle, src);
	}

	/**
	 * Retrieve the stored hash from the device.
	 * \param dst      - buff to store the hash in
	 * \return Number of bytes read from the device.
	 */
	size_t load_hash(Byte_range_ptr const &dst)
	{
		String<256> path { _path, "/", Services::HASH_STORAGE };
		Vfs::Vfs_handle* file_handle = _open(path, Vfs::Directory_service::OPEN_MODE_RDONLY);
		if (file_handle == nullptr) return 0;
		Vfs::Vfs_handle::Guard handle_guard(file_handle);

		return _read(file_handle, dst);
	}
};

#endif /* _TRUST_ANCHOR__CLIENT_H_ */
